//@author: vux
//@help: template for standard shaders
//@tags: template
//@credits: 

Texture2D Tex <string uiname="Texture";>;

SamplerState Samp 
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};
 
//transforms
float4x4 tW: WORLD;        //the models world matrix
float4x4 tV: VIEW;         //view matrix as set via Renderer (EX9)
float4x4 tP: PROJECTION;
float4x4 tWVP: WORLDVIEWPROJECTION;

float  scale <string uiname="Scale";> = (1.0);
float  cubicDistortion <string uiname="Cubic Distortion"; float uimin=0.0; float uimax=1.0;> = (0.0);
float  distortion <string uiname="Distortion"; float uimin=0.0; float uimax=1.0;> = (0.0);
float  chromaticDispersionRC <string uiname="Chromatic Dispersion Red/Cyan"; float uimin=0.0; float uimax=1.0;> = (0.0);
float  chromaticDispersionBY <string uiname="Chromatic Dispersion Blue/Yellow"; float uimin=0.0; float uimax=1.0;> = (0.0);
bool   invertDistort <string uiname="Invert Distortion"; float uimin=0.0; float uimax=1.0;> = (0.0);
float2 maxSize <string uiname="ScreenSize";> = float2(1.0,1.0);

float4x4 tTex <bool uvspace=true; string uiname="Texture Transform";>;

struct VS_IN
{
	float4 PosO : POSITION;
	float4 TexCd : TEXCOORD0;

};

struct vs2ps
{
    float4 Pos: SV_POSITION;
    float4 TexCd: TEXCOORD0;
};

vs2ps VS(VS_IN input)
{
    //declare output struct
    vs2ps Out;

    //transform position
    Out.Pos = mul(input.PosO, tWVP);
    
    //transform texturecoordinates
    Out.TexCd = mul(input.TexCd, tTex);

    return Out;
}

// Compute invert distortion with iterative approach
// (slower but more accurate rather than just inverting k & kcube)
float removeDistortion(float2 tex, float kcube, float f){
float r, rp, rlim, rplim, raw, err, slope;
bool bypass = false;
float ratio = maxSize.x/maxSize.y;
rp = sqrt(ratio * ratio * tex.x * tex.x + tex.y * tex.y);
if (kcube < 0.0) {
    rlim = sqrt((-1.0/3.0) / kcube);
    rplim = rlim * (1.0 + kcube*rlim*rlim);
    if (rp >= 0.99 * rplim) {
        f = rlim / rp;
        bypass = true;
    }
}

if(bypass == false){
    r = rp;
    for (int i = 0; i < 20; i++) {
        raw = kcube * r * r;
        slope = 1.0 + 3.0*raw;
        err = rp - r * (1.0 + raw);
        if (abs(err) < 1.0e-10) {
            i=20;
            }
        r += (err / slope);
        }
    f = r / rp;
    }

    return f;
}


/////////////////////////////


float4 PS(vs2ps In): SV_Target
{
    float ratio = maxSize.x/maxSize.y;
    float scaleFactor = 1.0/scale;

    //index of refraction of each color channel, causing chromatic dispersion
    float3 eta = float3(1.0+chromaticDispersionRC*0.09,
                        1.0+chromaticDispersionRC*0.06,
                        1.0+chromaticDispersionRC*0.03);
    float3 eta2 = float3(1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.03);

    float4 inputDistort;
    float2 tex = In.TexCd.xy/0.5 -1;
    
    // lens distortion coefficient
    float k = distortion;
    // cubic distortion value
    float kcube = cubicDistortion;
    float f = 0.0;

    if (invertDistort){
        //REMOVE DISTORTION
        f = removeDistortion(tex,kcube,f);

    }else{
        //APPLY DISTORTION      //*1.7 looks sooo wrong
        float r2 = ratio * ratio * (tex.x) * (tex.x) + (tex.y) * (tex.y);
        //only compute the cubic distortion if necessary
        if( kcube == 0.0){
                f = 1.0 + r2 * k;
        }else{
                f = 1.0 + r2 * (k + kcube * sqrt(r2));
        };
    }

    //Apply CA if necessary
    if (chromaticDispersionRC == 0.0 && chromaticDispersionBY == 0.0){
        float x = f*scaleFactor*(tex.x*0.5)+0.5;
        float y = f*scaleFactor*(tex.y*0.5)+0.5;
        inputDistort = Tex.Sample(Samp,float2(x,y));
    }else{
        float2 rCoords = (f*eta.r*eta2.r)*scaleFactor*(tex.xy*0.5)+0.5;
        float2 gCoords = (f*eta.g*eta2.g)*scaleFactor*(tex.xy*0.5)+0.5;
        float2 bCoords = (f*eta.b*eta2.b)*scaleFactor*(tex.xy*0.5)+0.5;
        float x = f*scaleFactor*(In.TexCd.x*0.5)+0.5;
        float y = f*scaleFactor*(In.TexCd.y*0.5)+0.5;
        inputDistort = float4(Tex.Sample(Samp,rCoords).r , Tex.Sample(Samp,gCoords).g ,
                              Tex.Sample(Samp,bCoords).b, Tex.Sample(Samp,float2(x,y)).a);
    }
    return inputDistort;
}

float4 PS2_noInvert(vs2ps In): SV_Target
{
    float ratio = maxSize.x/maxSize.y;
    float scaleFactor = 1.0/scale;

    //index of refraction of each color channel, causing chromatic dispersion
    float3 eta = float3(1.0+chromaticDispersionRC*0.09,
                        1.0+chromaticDispersionRC*0.06,
                        1.0+chromaticDispersionRC*0.03);
    float3 eta2 = float3(1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.03);

    float4 inputDistort;
    float2 tex = In.TexCd.xy;
    tex.x = 2.0*(tex.x/1 - 0.5);
    tex.y = 2.0*(tex.y/1 - 0.5);
    // lens distortion coefficient
    float k = distortion;
    // cubic distortion value
    float kcube = cubicDistortion;
    float f = 0.0;

    //APPLY DISTORTION
    float r2 = ratio * ratio * (tex.x) * (tex.x) + (tex.y) * (tex.y);
    //compute the cubic distortion
    f = 1.0 + r2 * (k + kcube * sqrt(r2));

    //Apply CA
    float2 rCoords = (f*eta.r*eta2.r)*scaleFactor*(tex.xy*0.5)+0.5;
    float2 gCoords = (f*eta.g*eta2.g)*scaleFactor*(tex.xy*0.5)+0.5;
    float2 bCoords = (f*eta.b*eta2.b)*scaleFactor*(tex.xy*0.5)+0.5;
    float x = f*scaleFactor*(In.TexCd.x*0.5)+0.5;
    float y = f*scaleFactor*(In.TexCd.y*0.5)+0.5;

    return float4(Tex.Sample(Samp,rCoords).r, Tex.Sample(Samp,gCoords).g,
                  Tex.Sample(Samp,bCoords).b, Tex.Sample(Samp,float2(x,y)).a);
}


float4 PS2_distortion(vs2ps In): SV_Target
{
    float ratio = maxSize.x/maxSize.y;
    float scaleFactor = 1.0/scale;

    //index of refraction of each color channel, causing chromatic dispersion
    float3 eta = float3(1.0+chromaticDispersionRC*0.09,
                        1.0+chromaticDispersionRC*0.06,
                        1.0+chromaticDispersionRC*0.03);
    float3 eta2 = float3(1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.03);

    float4 inputDistort;
    float2 tex = In.TexCd.xy;
    tex.x = 2.0*(tex.x/1 - 0.5);
    tex.y = 2.0*(tex.y/1 - 0.5);
    // lens distortion coefficient
    float k = distortion;
    // cubic distortion value
    float kcube = cubicDistortion;
    float f = 0.0;

    //APPLY DISTORTION
    float r2 = ratio * ratio * (tex.x) * (tex.x) + (tex.y) * (tex.y);
    //compute the cubic distortion
    f = 1.0 + r2 * (k + kcube * sqrt(r2));


    return float4(Tex.Sample(Samp, (f * eta * eta2)*scaleFactor*(tex.xy*0.5)+0.5));
}
float4 PS2_Inv_distortion(vs2ps In): SV_Target
{
    float ratio = maxSize.x/maxSize.y;
    float scaleFactor = 1.0/scale;

    //index of refraction of each color channel, causing chromatic dispersion
    float3 eta = float3(1.0+chromaticDispersionRC*0.09,
                        1.0+chromaticDispersionRC*0.06,
                        1.0+chromaticDispersionRC*0.03);
    float3 eta2 = float3(1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.03);

    float4 inputDistort;
    float2 tex = In.TexCd.xy;
    tex.x = 2.0*(tex.x/1 - 0.5);
    tex.y = 2.0*(tex.y/1 - 0.5);
    // lens distortion coefficient
    float k = distortion;
    // cubic distortion value
    float kcube = cubicDistortion;
    float f = 0.0;

    //APPLY DISTORTION
    float r2 = ratio * ratio * (tex.x) * (tex.x) + (tex.y) * (tex.y);
    //compute the cubic distortion
    f = 1.0 + r2 * (k + kcube * sqrt(r2));


    return float4(Tex.Sample(Samp, ((1/f) * eta * eta2)*scaleFactor*(tex.xy*0.5)+0.5));
}

float4 PS2_abberation(vs2ps In): SV_Target
{
    //Apply CA
    return float4(Tex.Sample(Samp,In.TexCd+chromaticDispersionRC*0.015).r,
                  Tex.Sample(Samp,In.TexCd+chromaticDispersionRC*0.0075).g,
                  Tex.Sample(Samp,In.TexCd+chromaticDispersionRC*0.00325).b,
                  Tex.Sample(Samp,In.TexCd).a);
}




technique10 Constant
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VS() ) );
		SetPixelShader( CompileShader( ps_4_0, PS() ) );
	}
}

technique10 ft_CubicLensDistortion_ps20_noInvert
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
		SetPixelShader( CompileShader( ps_4_0, PS2_noInvert() ) );
    }
}

technique10 ft_CubicLensDistortion_ps20_onlyDistortion
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
		SetPixelShader( CompileShader( ps_4_0,  PS2_distortion() ) );
    }
}

technique10 ft_CubicLensDistortion_ps20_onlyCYAbberation
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
		SetPixelShader( CompileShader( ps_4_0, PS2_abberation() ) );
    }
}

technique10 ft_Inv_CubicLensDistortion_ps20_onlyDistortion
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
		SetPixelShader( CompileShader( ps_4_0, PS2_Inv_distortion() ) );
    }
}



